/****************************************************************************/
/* Copyright 2000 Compaq Computer Corporation.                              */
/*                                           .                              */
/* Copying or modifying this code for any purpose is permitted,             */
/* provided that this copyright notice is preserved in its entirety         */
/* in all copies or modifications.  COMPAQ COMPUTER CORPORATION             */
/* MAKES NO WARRANTIES, EXPRESSED OR IMPLIED, AS TO THE USEFULNESS          */
/* OR CORRECTNESS OF THIS CODE OR ITS FITNESS FOR ANY PARTICULAR            */
/* PURPOSE.                                                                 */
/****************************************************************************/
/*
 * bootldr.h
 */
#ifndef _BOOTLDR_H_
#define _BOOTLDR_H_

/* defined for release candidates to disable certain functions */
#define RELEASE

/* include assembler constants and boot configuration */
#include "regs-21285.h"
#include "bootconfig.h"

/* basic data types and constants */
#ifdef __linux__
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
#endif
typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned int   dword;
typedef void *vaddr_t;
typedef unsigned long paddr_t;

typedef long size_t;

typedef void (*vfuncp)(void);

#define CTL_CH(c)	((c) - 'a' + 1)
#define CTL_REG_READ(addr)	    (*(volatile unsigned long *)(addr))
#define CTL_REG_WRITE(addr, val)    (*(volatile unsigned long *)(addr) = (val))

#define CTL_REG_READ_BYTE(addr)	    (*(volatile unsigned char *)(addr))
#define CTL_REG_WRITE_BYTE(addr, val)    (*(volatile unsigned char *)(addr) = (val))

#define TRUE           1
#define FALSE          0
#define NULL           ((void *) 0)

/* bootloader delay time */
#define TIMEOUT        350000

#define MAX_BOOTARGS_LEN 2048
/* 21285 CSR Access Macros - use these to peek/poke various 21285 registers */
#define CSR_READ_BYTE(p)		(*(volatile byte *)(CSR_BASE+p))
#define CSR_WRITE_BYTE(p, c)		(*(volatile byte *)(CSR_BASE+p) = c)

/* 21285 PCI memory maps - expected to be configured by NetBSD kernel */
#define DC21285_ARMCSR_VBASE		0xF4000000
#define	DC21285_ARMCSR_VSIZE		0x00100000	/* 1MB */
#define	DC21285_CACHE_FLUSH_VBASE	0xF4100000
#define	DC21285_CACHE_FLUSH_VSIZE	0x00100000	/* 1MB */
#define	DC21285_PCI_IO_VBASE		0xF4200000
#define	DC21285_PCI_IO_VSIZE		0x00100000	/* 1MB */
#define	DC21285_PCI_IACK_VBASE		0xF4300000
#define	DC21285_PCI_IACK_VSIZE		0x00100000	/* 1MB */
#define	DC21285_PCI_TYPE_1_CONFIG_VBASE	0xF5000000
#define	DC21285_PCI_TYPE_1_CONFIG_VSIZE	0x01000000	/* 16MB */
#define	DC21285_PCI_TYPE_0_CONFIG_VBASE	0xF6000000
#define	DC21285_PCI_TYPE_0_CONFIG_VSIZE	0x01000000	/* 16MB */
#define DC21285_PCI_MEM_VBASE		0xF7000000
#define DC21285_PCI_MEM_VSIZE		0x09000000	/* 144MB */

/* MMU Level 1 Page Table Constants */
#define MMU_FULL_ACCESS (3 << 10)
#define MMU_DOMAIN      (0 << 5)
#define MMU_SPECIAL     (0 << 4)
#define MMU_CACHEABLE   (1 << 3)
#define MMU_BUFFERABLE  (1 << 2)
#define MMU_SECTION     (2)
#define MMU_SECDESC  (MMU_FULL_ACCESS | MMU_DOMAIN | \
                      MMU_SPECIAL | MMU_SECTION)

extern unsigned long *mmu_table;

/*
 * Points to the beginning of where the bootldr is in memory (beginning of flash, beginning of dram, etc.)
 * For position independence
 */
extern long bootldr_start;


/*
flash memory region map. This basically splits the bootblocks of the
flash as folows:

32K - bootloader
32K - flash variables/data
*/

#if DeadCode
#define BOOTLDR_SECTOR          0 /* and 1 and 2 too */
#define BOOTLDR_OFFSET          0
#define BOOTLDR_SIZE            SZ_48K
#define BOOTLDR_START           (FLASH_BASE+BOOTLDR_OFFSET)
#define BOOTLDR_END             (BOOTLDR_START+BOOTLDR_SIZE)
#define VBOOTLDR_START          (UNCACHED_FLASH_BASE+BOOTLDR_OFFSET)
#define VBOOTLDR_END            (VBOOTLDR_START+BOOTLDR_SIZE)
#endif

enum bootblk_flags {
    BB_NONE = 0,
    BB_RUN_FROM_RAM = 1
};

struct bootblk_command {
    const char *cmdstr;
    void (*cmdfunc)(int argc, const char **);
    const char *helpstr;
    enum bootblk_flags flags;
};

#define DEFAULT_XMODEM_BLOCK_SIZE 1024            /* default size of transmit blocks */
#define MAX_XMODEM_BLOCK_SIZE	1024              /* maximum size of transmit blocks */
extern long xmodem_block_size;

/* must match definition in linux/include/asm/mach-types.h */
#define MACH_TYPE_CATS 6
#define MACH_TYPE_PERSONAL_SERVER 17
#define MACH_TYPE_SA1100 16
#define MACH_TYPE_H3600                22
#define MACH_TYPE_H3100                136
#define MACH_TYPE_H3800                137
#define MACH_TYPE_H3600_ASCII          "22"
#define MACH_TYPE_H3100_ASCII          "136"
#define MACH_TYPE_H3800_ASCII          "137"
#define MACH_TYPE_ASSABET 25
#define MACH_TYPE_JORNADA720 48
#define MACH_TYPE_OMNIMETER 49
#define MACH_TYPE_SPOT 53
#define MACH_TYPE_GATOR 103
#define MACH_TYPE_JORNADA56X 167

#ifdef CONFIG_ASSABET
#define MACH_TYPE MACH_TYPE_ASSABET
#endif
#ifdef CONFIG_BITSY
#define MACH_TYPE MACH_TYPE_H3600
#endif
#ifdef CONFIG_JORNADA720
#define MACH_TYPE MACH_TYPE_JORNADA720
#endif
#ifdef CONFIG_JORNADA56X
#define MACH_TYPE MACH_TYPE_JORNADA56X
#endif
#ifdef CONFIG_OMNIMETER
#define MACH_TYPE MACH_TYPE_OMNIMETER
#endif
#ifdef CONFIG_SKIFF
#define MACH_TYPE MACH_TYPE_PERSONAL_SERVER
#endif
#ifdef CONFIG_SPOT
#define MACH_TYPE MACH_TYPE_SPOT
#endif
#ifdef CONFIG_GATOR
#define MACH_TYPE MACH_TYPE_GATOR
#endif

void initialize_by_mach_type(void);

/* External Functions - written in assembler [boot.s] */
void boot(void *bootinfo,long startaddr);
void bootLinux(void *bootinfo, long machine_type, long startaddr);

#ifdef NoLibC
void *memcpy(char *dst, const char *src, long n);
int  memcmp(const char *buf1,const char *buf2, long n);
#endif
unsigned long strtoul(const char *str, char **endptr, int requestedbase);
char *strchr(const char *str, char c);
char *strrchr(const char *str, char c);
char *strcat(char *dst, const char *str);
int strcmp(const char *s1, const char *s2);
int strncmp(const char *s1, const char *s2, int len);
int memmem(const char *haystack,int num1, const char *needle,int num2);
unsigned long strtoul(const char *str, char **endptr, int requestedbase);

void PrintChar(char c, void *uart); 
void putLabeledWord(const char *msg, unsigned long value);
void putHexInt32(unsigned long value);
void putHexInt16(word value);
void putHexInt8(byte value);
void delay_seconds(unsigned long seconds);
void binary_dld(unsigned long img_size, unsigned long img_dest);
void dwordtodecimal(char *buf, unsigned long x);
void hex_dump(unsigned char *data, size_t num);

long get_system_rev();
void program_all_eeprom();

long enable_caches(int dcache_enabled, int icache_enabled);
void flush_caches(void);
void bootConfigureMMU(void);
void enableMMU(void);
void flashConfigureMMU(unsigned long flash_size);


/* asm routines */
void flushTLB(void);
int asmEnableCaches(int dcache_enabled, int icache_enabled);
void writeBackDcache(unsigned long);
unsigned long readPC(void);
unsigned long readCPR0(void);
unsigned long readCPR1(void);
unsigned long readCPR2(void);
unsigned long readCPR3(void);
unsigned long readCPR4(void);
unsigned long readCPR5(void);
unsigned long readCPR6(void);
unsigned long readCPR13(void);
unsigned long readCPR14(void);



extern void
let_uart_drain(
    unsigned long status_addr);

#define	let_uart_drain0()    let_uart_drain(SA1100_UART3_UTSR1)

extern void
exec_string(
    char*   buf);

extern int
push_cmd_chars(
    char*   chars,
    int	    len);

extern void set_egpio(int mask);
extern void clr_egpio(int mask);

#include "boot_flags.h"

#define	BL_ISPRINT(ch)	    (((ch)>=' ') && ((ch) < 128))

extern int
reboot_button_is_enabled(
    void);

extern void
enable_reboot_button(
    void);

extern void
disable_reboot_button(
    void);

extern void
bootldr_goto_sleep(
    void*   pspr);

extern void
print_mem_size(
    char*   hdr,
    long    mem_size);

extern unsigned char amRunningFromRam(void);
#if defined(CONFIG_LOAD_KERNEL)
extern void boot_kernel(const char *root_filesystem_name, 
                        vaddr_t kernel_region_start, size_t kernel_region_size, int argc, const char **argv, int no_copy);
#endif

#endif /* _BOOTLDR_H_ */

